
/* S2 cpu micro architecture simulator

   5 December 2001
   s21  18 Jan 2007

   modify from s1 sim

   Computer Architecture 1997, 1998

   P. Chongstitvatana
   Department of Computer Engineering
   Chulalongkorn University

*/

#include "s21.h"

int R[32], Ir, Mdr, Mar, Pc, T, Z, S, M[MAXMEM];
int runflag, savePc, freecell;
extern char *cp2;		// output buffer

void ifetch(void){      // instruction fetch
	Mar = Pc;
	Mdr = M[Mar]; Pc = Pc + 1;
}

void decode(void){
	Ir = Mdr;
}

/* alu operation input : a, b
 affect condition code Z,S
 carry and overflow are not implemented */

int alu(int op, int a, int b){
  int t;
  switch(op) {
  case ADD: t = a + b; break;
  case SUB: t = a - b; break;
  case MUL: t = a * b; break;
  case DIV: t = a / b; break;
  case AND: t = a & b; break;
  case OR:  t = a | b; break;
  case XOR: t = a ^ b; break;
  case SHL: t = a << b; break;
  case SHR: t = a >> b; break;
  case EQ: t = a == b; break;
  case NE: t = a != b; break;
  case LT: t = a < b; break;
  case LE: t = a <= b; break;
  case GT: t = a > b; break;
  case GE: t = a >= b; break;
  }
  if( t == 0 ) Z = 1; else Z = 0;
  if( t < 0 )  S = 1; else S = 0;
  return t;
}

int adder(int a, int b){
	return a + b;
}
/*
int testCc(int cond ){	// test condition code
  switch( cond ){
  case ALWAYS: return 1;
  case EQ: return Z;
  case NEQ: return Z == 0;
  case LT: return S && (Z == 0);
  case LE: return S || Z;
  case GE: return (S == 0) || Z;
  case GT: return (S == 0) && (Z == 0);
  default: error("undefine conditional code");
  }
  return 0;
}
*/
// write register, check for R[0]
void wR(int rn, int value){
	if(rn != 0) R[rn] = value;
}

void opRR(int op){
	T = alu(op,R[IRr2()],R[IRr3()]);
	wR(IRr1(),T);
}

void opRI(int op){
	T = alu(op,R[IRr2()],IRdisp());
	wR(IRr1(),T);
}

int signx2( int d ){	// sign bit 21 extended
	if( d & 0x0200000 ) return d | 0xFFC00000;
	return d;
}

PRIVATE int car(int a){ return M[a]; }
PRIVATE int cdr(int a){ return M[a+1]; }
PRIVATE int isatom(int a){ return M[a] < 10; }

PRIVATE void prlist(int a){
	if( a == NIL ) return;
	if( isatom(a) )
		cp2 += sprintf(cp2,"%d ",M[a+1]);
	else{
		cp2 += sprintf(cp2,"(");
		while( a != NIL ){
			if(car(a) == NIL) cp2 += sprintf(cp2,"() ");
			else prlist(car(a));
			a = cdr(a);
		}
		cp2 += sprintf(cp2,")");
	}
}

// list memory management
// cell area is M[10000...]
// freecell pointed to available cell

void trap( int num ){	// special function
	int n;
	char buf[80];
	switch( num ) {
	case 0:  // stop
//		report();
		printf("stop, execute %d instructions\n",ninst);
		runflag = 0;
		break;
	case 1: // print integer
		n = sprintf(cp2,"%d", R[30]);
		cp2 += n;
		printf("\n<%d>\n",R[30]);
		break;
	case 2: // printc
		n = sprintf(cp2,"%c", R[30]);
		cp2 += n;
		break;
	case 3: // input
		scanf("%s",buf);
//		printf("<%s>\n",buf);
		if(isdigit(buf[0])){
			R[30] = 1;
			R[29] = atoi(buf);
		}else{
			R[30] = 2;
			R[29] = (int)(buf[0]);
		}
		break;
	case 5: // get a cell (for list)
		if (freecell >= MAXMEM) error("memory overflow");
		R[28] = freecell;
		freecell += 2;
		break;
	case 6: // print list
//		printf("freecell %d\n",freecell);
		prlist(R[30]);
		break;
	}
}

void run(void){		// execute one instruction
	savePc = Pc;
	ifetch();
	decode();
//	countclock();
	ninst++;
	switch( IRop() ) {
	case LDA:
		Mar = IRads();
		Mdr = M[Mar];
		wR(IRr1(),Mdr);
		break;
	case LDD:
		Mar = adder(R[IRr2()],IRdisp());  // not affect flag
		Mdr = M[Mar];
		wR(IRr1(),Mdr);
		break;
	case STA:
		Mar = IRads();
		Mdr = R[IRr1()];
		M[Mar] = Mdr;
		break;
	case STD:
		Mar = adder(R[IRr2()],IRdisp());
		Mdr = R[IRr1()];
		M[Mar] = Mdr;
		break;
	case MVI:
		wR(IRr1(),signx2(IRads()));
		break;
	case JMP:
//		if( testCc(IRr1()) )
			Pc = IRads();
		break;
	case JAL:
		wR(IRr1(),Pc);
		Pc = IRads();
		break;
	case JT:
		if( R[IRr1()] != 0 ) Pc = IRads();
		break;
	case JF:
		if( R[IRr1()] == 0 ) Pc = IRads();
		break;
	case ADDI: opRI(ADD); break;
	case SUBI: opRI(SUB); break;
	case MULI: opRI(MUL); break;
	case DIVI: opRI(DIV); break;
	case ANDI: opRI(AND); break;
	case ORI:  opRI(OR); break;
	case XORI: opRI(XOR); break;
	case EQI: opRI(EQ); break;
	case NEI: opRI(NE); break;
	case LTI: opRI(LT); break;
	case LEI: opRI(LE); break;
	case GTI: opRI(GT); break;
	case GEI:  opRI(GE); break;
	case SHLI: opRI(SHL); break;
	case SHRI: opRI(SHR); break;
	case XOP: xop(); break;   /* extended op */
	default: error("undefine op");
	}
}

void xop(void){
	switch( IRxop() ) {
	case ADD:
	case SUB:
	case MUL:
	case DIV:
	case AND:
	case OR:
	case XOR:
	case EQ:
	case NE:
	case LT:
	case LE:
	case GT:
	case GE:
	case SHL:
	case SHR: opRR(IRxop()); break;
	case MV: wR(IRr1(),R[IRr2()]); break;
	case LDX:
		Mar = adder(R[IRr2()],R[IRr3()]); // not affect flag
		Mdr = M[Mar];
		wR(IRr1(),Mdr);
		break;
	case STX:
		Mar = adder(R[IRr2()],R[IRr3()]);
		Mdr = R[IRr1()];
		M[Mar] = Mdr;
		break;
	case RET: Pc = R[IRr1()]; break;
	case TRAP: trap(IRr1()); break;
	case PUSH:
		wR(IRr1(),R[IRr1()]+1);
		Mar = R[IRr1()];
		Mdr = R[IRr2()];
		M[Mar] = Mdr;
		break;
	case POP:
		Mar = R[IRr1()];
		Mdr = M[Mar];
		wR(IRr2(),Mdr);
		wR(IRr1(),R[IRr1()]-1);
		break;
	case NOT:
		if(R[IRr2()] != 0) wR(R[IRr1()],0);
		else wR(R[IRr1()],1);
		break;
	default: error("undefine op");
	}
}

int main(int argc, char *argv[]){
	if( argc < 2 ) {
		printf("usage : s21 objfile\n");
		exit(0);
	}
	loadprogram(argv[1]);
/*
	if(argc > 2) {
		printf("obj %s\n",argv[2]);
		loadobj(argv[2]);
	}
*/
	clock = 0;
	ninst = 0;
	Pc = 0;
	Z = 0;
	S = 0;
	R[0] = 0;
	freecell = 10000;
	runflag = 1;
	interp();
	return 0;
}

