
/* S2 cpu micro architecture simulator

   5 December 2001
   s21  18 Jan 2007

   modify from s1 sim

   Computer Architecture 1997, 1998

   simplify (direct) for calculator project   29 Jan 2016
   no Z, S

   P. Chongstitvatana
   Department of Computer Engineering
   Chulalongkorn University

*/

#include "s21.h"

int R[32], Ir, Pc, M[MAXMEM];
int runflag, savePc;
extern char *cp2;		// output buffer


int signx2( int d ){	// sign bit 21 extended
	if( d & 0x0200000 ) return d | 0xFFC00000;
	return d;
}

void trap( int num ){	// special function
	int n;
	char buf[80];
	switch( num ) {
	case 0:  // stop
//		report();
		printf("stop, execute %d instructions\n",ninst);
		runflag = 0;
		break;
	case 1: // print integer
		n = sprintf(cp2,"%d", R[30]);
		cp2 += n;
		printf("\n<%d>\n",R[30]);
		break;
	case 2: // printc
		n = sprintf(cp2,"%c", R[30]);
		cp2 += n;
		printf("%c",R[30]);
		break;
	case 3: // input  for calculator project
		scanf("%s",buf);
//		printf("<%s>\n",buf);
		if(isdigit(buf[0])){
			R[30] = 1;
			R[29] = atoi(buf);
		}else{
			R[30] = 2;
			R[29] = (int)(buf[0]);
		}
		break;
	case 4: // dump registers
		dumpreg();
		break;
	}
}

void run(void){		// execute one instruction
	int op, ads, d, r1, r2, r3;

	savePc = Pc;
	Ir = M[Pc];		// instruction fetch
	Pc++;
	op = IRop();	// decode
	ads = IRads();
	d = IRdisp();	// signx  17-bit
	r1 = IRr1();
	r2 = IRr2();
	r3 = IRr3();
//  countclock();
	ninst++;
	R[0] = 0;		// always 0
	switch( op ) {
	case LDA:  R[r1] = M[ads]; break;
	case LDD:  R[r1] = M[R[r2]+d]; break;
	case STA:  M[ads] = R[r1];  break;
	case STD:  M[R[r2]+d] = R[r1]; break;
	case MVI:  R[r1] = signx2(ads); break;
	case JMP:  Pc = ads; break;
	case JAL:
		R[r1] = Pc;
		Pc = ads;
		break;
	case JT: if( R[r1] != 0 ) Pc = ads; break;
	case JF: if( R[r1] == 0 ) Pc = ads; break;
	case ADDI: R[r1] = R[r2] + d; break;
	case SUBI: R[r1] = R[r2] - d; break;
	case MULI: R[r1] = R[r2] * d; break;
	case DIVI: R[r1] = R[r2] / d; break;
	case ANDI: R[r1] = R[r2] & d; break;
	case ORI:  R[r1] = R[r2] | d; break;
	case XORI: R[r1] = R[r2] ^ d; break;
	case EQI:  R[r1] = R[r2] == d; break;
	case NEI:  R[r1] = R[r2] != d; break;
	case LTI:  R[r1] = R[r2] <  d; break;
	case LEI:  R[r1] = R[r2] <= d; break;
	case GTI:  R[r1] = R[r2] >  d; break;
	case GEI:  R[r1] = R[r2] >= d; break;
	case SHLI: R[r1] = R[r2] << d; break;
	case SHRI: R[r1] = R[r2] >> d; break;
	case XOP: xop(); break;
	default: error("undefine op");
	}
}

void xop(void){
	int op2, r1, r2, r3;
	op2 = IRxop();
	r1 = IRr1();
	r2 = IRr2();
	r3 = IRr3();

	switch( op2 ){
	case ADD: R[r1] = R[r2] + R[r3]; break;
	case SUB: R[r1] = R[r2] - R[r3]; break;
	case MUL: R[r1] = R[r2] * R[r3]; break;
	case DIV: R[r1] = R[r2] / R[r3]; break;
	case AND: R[r1] = R[r2] & R[r3]; break;
	case OR:  R[r1] = R[r2] | R[r3]; break;
	case XOR: R[r1] = R[r2] ^ R[r3]; break;
	case EQ:  R[r1] = R[r2] == R[r3]; break;
	case NE:  R[r1] = R[r2] != R[r3]; break;
	case LT:  R[r1] = R[r2] <  R[r3]; break;
	case LE:  R[r1] = R[r2] <= R[r3]; break;
	case GT:  R[r1] = R[r2] >  R[r3]; break;
	case GE:  R[r1] = R[r2] >= R[r3]; break;
	case SHL: R[r1] = R[r2] << R[r3]; break;
	case SHR: R[r1] = R[r2] >> R[r3]; break;
	case MV:   R[r1] = R[r2]; break;
	case LDX:  R[r1] = M[R[r2] + R[r3]]; break;
	case STX:  M[R[r2] + R[r3]] = R[r1]; break;
	case RET:  Pc = R[r1]; break;
	case TRAP: trap(r1); break;
	case PUSH:
		R[r1]++;			// increment stack pointer
		M[R[r1]] = R[r2];
		break;
	case POP:
		R[r2] = M[R[r1]];
		R[r1]--;			// decrement stack pointer
		break;
	case NOT:
		R[r1] = (R[r2] == 0) ? ~0 : 0;  break;
	default:
		error("undefine op");
	}
}

int main(int argc, char *argv[]){
	if( argc < 2 ) {
		printf("usage : s21 objfile\n");
		exit(0);
	}
	loadprogram(argv[1]);
/*
	if(argc > 2) {
		printf("obj %s\n",argv[2]);
		loadobj(argv[2]);
	}
*/
	clock = 0;
	ninst = 0;
	Pc = 0;

	runflag = 1;
	interp();
	return 0;
}

